from .i18n import _
from .error import DictConsistencyError
from .config import Config


class Path:
    """Helper class to handle the `path` attribute of a CreoleObjSpace
    instance.

    sample: path="creole.general.condition"
    """
    def __init__(self):
        self.variables = {}
        self.families = {}
        self.full_paths_families = {}
        self.full_paths_variables = {}

    # Family
    def add_family(self,
                   namespace: str,
                   name: str,
                   variableobj: str,
                   ) -> str:  # pylint: disable=C0111
        if namespace == Config['variable_namespace']:
            full_name = '.'.join([namespace, name])
            self.full_paths_families[name] = full_name
        else:
            if '.' not in name:
                raise DictConsistencyError(_(f'Variable "{name}" in namespace "{namespace}" must have dot'), 39)
            full_name = name
        if full_name in self.families and \
                self.families[full_name]['variableobj'] != variableobj:  # pragma: no cover
            raise DictConsistencyError(_(f'Duplicate family name "{name}"'), 37)
        self.families[full_name] = dict(name=name,
                                        namespace=namespace,
                                        variableobj=variableobj,
                                        )
        variableobj.path = full_name

    def add_leadership(self,
                       namespace: str,
                       path: str,
                       variableobj: str,
                       ) -> str:  # pylint: disable=C0111
        self.families[path] = dict(name=path,
                                   namespace=namespace,
                                   variableobj=variableobj,
                                   )
        variableobj.path = path

    def _get_family(self,
                    name: str,
                    namespace: str=None,
                    ):
        # if main namespace, get full_path
        if '.' not in name and namespace in [None, Config['variable_namespace']] and name in self.full_paths_families:
            name = self.full_paths_families[name]
        dico = self.families[name]
        if namespace and dico['namespace'] != Config['variable_namespace'] and namespace != dico['namespace']:
            raise DictConsistencyError(_(f'A family located in the "{dico["namespace"]}" namespace shall not be used in the "{namespace}" namespace'), 38)
        return dico

    def get_family(self,
                   name: str,
                   namespace: str,
                   ) -> 'Family':  # pylint: disable=C0111
        return self._get_family(name, namespace)['variableobj']

    def family_is_defined(self,
                          name: str,
                          namespace: str,
                          ) -> str:  # pylint: disable=C0111
        try:
            self._get_family(name, namespace)
            return True
        except KeyError:
            return False

    # Leadership
    def set_leader(self,
                   namespace: str,
                   leader_family_name: str,
                   leadership_name: str,
                   name: str,
                   ) -> None:  # pylint: disable=C0111
        # need rebuild path and move object in new path
        old_path = leader_family_name + '.' + name
        leadership_path = leader_family_name + '.' + leadership_name
        new_path = leadership_path + '.' + name
        self.variables[new_path] = self.variables.pop(old_path)
        self.variables[new_path]['leader'] = leadership_path
        self.variables[new_path]['variableobj'].path = new_path
        self.variables[new_path]['family'] = leadership_path
        if namespace == Config['variable_namespace']:
            self.full_paths_variables[name] = new_path

    def is_in_leadership(self, name):
        return self._get_variable(name)['leader'] is not None

    def is_leader(self, path):  # pylint: disable=C0111
        variable = self._get_variable(path)
        if not variable['leader']:
            return False
        leadership = self.get_family(variable['leader'], variable['variableobj'].namespace)
        return leadership.variable[0].path == path

    # Variable
    def add_variable(self,
                     namespace: str,
                     name: str,
                     family: str,
                     is_dynamic: bool,
                     variableobj,
                     ) -> str:  # pylint: disable=C0111
        if '.' not in name:
            full_path = '.'.join([family, name])
            if namespace == Config['variable_namespace']:
                self.full_paths_variables[name] = full_path
        else:
            full_path = name
        variableobj.path = full_path
        self.variables[full_path] = dict(name=name,
                                         family=family,
                                         leader=None,
                                         is_dynamic=is_dynamic,
                                         variableobj=variableobj,
                                         )

    def get_variable_obj(self,
                         name: str,
                         ) -> 'Variable':  # pylint: disable=C0111
        variable, suffix = self._get_variable(name, with_suffix=True)
        if suffix:
            raise DictConsistencyError(_(f"{name} is a dynamic variable"), 36)
        return variable['variableobj']

    def get_variable_family_path(self,
                                 name: str,
                                 ) -> str:  # pylint: disable=C0111
        return self._get_variable(name)['family']

    def get_variable_path(self,
                          name: str,
                          current_namespace: str,
                          allow_variable_namespace: str=False,
                          ) -> str:  # pylint: disable=C0111
        dico, suffix = self._get_variable(name,
                                          with_suffix=True,
                                          )
        namespace = dico['variableobj'].namespace
        if not allow_variable_namespace and namespace not in [Config['variable_namespace'], 'services'] and current_namespace != namespace:
            raise DictConsistencyError(_(f'A variable located in the "{namespace}" namespace shall not be used in the "{current_namespace}" namespace'), 41)
        return dico['variableobj'].path, suffix

    def path_is_defined(self,
                        name: str,
                        ) -> str:  # pylint: disable=C0111
        if '.' not in name and name not in self.variables and name in self.full_paths_variables:
            return True
        return name in self.variables

    def variable_is_dynamic(self,
                            name: str,
                            ) -> bool:
        return self._get_variable(name)['is_dynamic']

    def _get_variable(self,
                      name: str,
                      with_suffix: bool=False,
                      ) -> str:
        if name not in self.variables:
            if '.' not in name and name in self.full_paths_variables:
                name = self.full_paths_variables[name]
            elif with_suffix:
                for var_name, full_path in self.full_paths_variables.items():
                    if name.startswith(var_name):
                        variable = self._get_variable(full_path)
                        if variable['is_dynamic']:
                            return variable, name[len(var_name):]
        if name not in self.variables:
            raise DictConsistencyError(_('unknown option {}').format(name), 42)
        if with_suffix:
            return self.variables[name], None
        return self.variables[name]
